!***********************************************************************
!      RAISHIN code 3-dimensional Resistive Special Relativistic MHD version 
!      main program
!      written by Y. Mizuno
!***********************************************************************
!
program main

  use pram
  implicit none
  include 'mpif.h' ! for MPI
!
!======================================================================@
!    Difinition for variables
!======================================================================@
!
! uu: converved variables (cell-center)
! uo: 1 step before data of uu
! u0: initial value of uu
! uusr: store data for restart
! uu*r, uu*l: conserved variables (cell-surface)
!
  real(8), allocatable :: uu(:,:,:,:), uh(:,:,:,:), uo(:,:,:,:), &
          u0(:,:,:,:), uusr(:,:,:,:), uuir(:,:,:,:), uuil(:,:,:,:), &
          uujr(:,:,:,:), uujl(:,:,:,:), uukr(:,:,:,:), uukl(:,:,:,:)
!
! uri: primitive variables (cell-center)
! urio: 1 step before data of uri
! uri0: initial value of uri
! urisr: store data for restart
! uri*r, uri*l: premitive variables (cell-surface)
!
  real(8), allocatable :: uri(:,:,:,:), urio(:,:,:,:), &
          uri0(:,:,:,:), urisr(:,:,:,:), &
          uriir(:,:,:,:), urijr(:,:,:,:), urikr(:,:,:,:), &
          uriil(:,:,:,:), urijl(:,:,:,:), urikl(:,:,:,:)
!
! Metric terms: hh,hi,ho: cell-center
!               h*i,h*j,h*k: cell-boundary
!
  real(8), allocatable :: hh(:,:,:,:), &
          hhi(:,:,:,:), hhj(:,:,:,:), hhk(:,:,:,:), &
          hi(:,:,:,:,:), hii(:,:,:,:), hij(:,:,:,:), hik(:,:,:,:), &
          ho(:,:,:,:,:), hoi(:,:,:,:), hoj(:,:,:,:), hok(:,:,:,:) 
  real(8), allocatable :: gm(:,:,:,:,:)
!
! ww: numerical flux (cell-center)
! wwo: 1step before numerical flux (cell-center) 
! ww*r, ww*l: numerical flux (cell-boundary)
!
  real(8), allocatable :: ww(:,:,:,:,:), wwo(:,:,:,:,:), &
          wwir(:,:,:,:), wwjr(:,:,:,:), wwkr(:,:,:,:), & 
          wwil(:,:,:,:), wwjl(:,:,:,:), wwkl(:,:,:,:)
!
! sf: source term
!
  real(8), allocatable :: sf(:,:,:,:)
!
! cmax, cmin: maximum and minimum wave speed
!
  real(8), allocatable :: cmaxi(:,:,:), cmini(:,:,:), &
          cmaxj(:,:,:), cminj(:,:,:), cmaxk(:,:,:), cmink(:,:,:)
!
! Primitive variables
!
  real(8), allocatable :: b1(:,:,:), b2(:,:,:), b3(:,:,:), &
          v1(:,:,:), v2(:,:,:), v3(:,:,:), de(:,:,:), pr(:,:,:), &
          e1(:,:,:), e2(:,:,:), e3(:,:,:), ch(:,:,:), sg(:,:,:)
!
! for GLM, 
!
  real(8), allocatable :: ps(:,:,:), ph(:,:,:)
  real(8) :: glmch, glmcp
!
! Grid: x1: cell-center, x1a: cell-boundary
!       dx1: x1(i+1)-x1(i), dx1b=x1a(i+1)-x1a(i)
!       akap1=dt/dx1, akap1b=dt/dx1b
!
  real(8) :: x1(imax), x2(jmax), x3(kmax), &
             x1a(imax), x2a(jmax), x3a(kmax), &
             dx1(imax), dx2(jmax), dx3(kmax), &
             dx1b(imax-1), dx2b(jmax-1), dx3b(kmax-1), &
             akap1(imax), akap2(jmax), akap3(kmax), &
             akap1b(imax-1), akap2b(jmax-1), akap3b(kmax-1)
!
!     for Kolmogorov-like Power-law Spectrum 
!
  real(8) :: wkn(nkmax), pkn1d(nkmax), pkn2d(nkmax), &
             pkn3d(nkmax), thetan(nkmax), phin(nkmax), &
             thetan1(nkmax), phin1(nkmax)
!
  character*256 :: filename, filename1
  integer :: i, j, k, m, n 
  integer :: is, ie, is1, ie1
  integer :: js, je, js1, je1
  integer :: ks, ke, ks1, ke1 
  integer :: iup, idown, jup, jdown, kup, kdown
  integer :: icputable(-1:iprocs,-1:jprocs,-1:kprocs)
  real(8) :: dt, dtshot, dt1, dt2, dt3, dtcfl, dtcflg 
  real(8) :: dx1a, dx2a, dx3a, akap1a, akap2a, akap3a
  real(8) :: time, timeh, tnext, timesr, timeo, time1
  integer :: istop, ierror
  integer :: it0, it, ihi, its, itsr, ihisr, isr, iit
  integer :: ig, nd
  real(8) :: cputime
  integer :: nm0
  real(8) :: cl, amdamax, amdamin, totpkn1d, totpkn2d, totpkn3d
  real(8) :: e0, rj, vj, etaj, betaj, cmac, bm
  integer :: imagj
  real(8) :: vpre1, vplu1, opre1, omk1
  real(8) :: cuc, eta0
  integer :: npe, myrank, merr, myranki, myrankj, myrankk
!
!======================================================================@
! for MPI
  call mpi_init(merr)
  call mpi_comm_size(mpi_comm_world,npe   ,merr)
  call mpi_comm_rank(mpi_comm_world,myrank,merr)
!
  if(npe .ne. iprocs*jprocs*kprocs) then
    if(myrank .eq. 0) then
      write(*,*) 'Error: for ncpu \= iprocs*jprocs*kprocs'
    endif
    call mpi_finalize(merr)
    stop
  endif 
!
  call makecputable(myrank,icputable,myranki,myrankj,myrankk)
  call pararange(1,imax,iprocs,myranki,is,ie)
  call pararange(1,jmax,jprocs,myrankj,js,je)
  call pararange(1,kmax,kprocs,myrankk,ks,ke)
!      
!    Buffer grid number
!
  if(myrank .eq. 0) then
    write(*,*) 'npe,iprocs,jprocs,kprocs=', npe, iprocs, jprocs, kprocs
  endif
  write(*,*) 'myrank,myranki,myrankj,myrankk=', myrank, myranki, &
             myrankj, myrankk

  nm0=1
      
  if(irec .le. 3 .or. irec .eq. 11) then
    nm0=2
  elseif(irec .eq. 4 .or. irec .ge. 6 .and. irec .le. 10) then
    nm0=3
  elseif(irec .eq. 5) then
    nm0=4
  endif
!
  if(myranki .eq. 0) then
    is1=1
  else
    is1=is-nm0
  endif
  if(myranki .eq. iprocs-1) then
    ie1=ie
  else
    ie1=ie+nm0
  endif
!
  if(myrankj .eq. 0) then
    js1=1
  else
    js1=js-nm0
  endif
  if(myrankj .eq. jprocs-1) then
    je1=je
  else
    je1=je+nm0
  endif
!
  if(myrankk .eq. 0) then
    ks1=1
  else
    ks1=ks-nm0
  endif
  if(myrankk .eq. kprocs-1) then
    ke1=ke
  else
    ke1=ke+nm0
  endif
!
  if(ie1-is1+1 .le. 2*nm0 .or. je1-js1+1 .le. 2*nm0 &
     .or. ke1-ks1+1 .le. 2*nm0) then
    write(*,*) 'Error: for grid size < minimum number of grid'
    write(*,*) 'myrank, ith grid #, jth grid #, kth grid # =',myrank, &
               ie1-is1+1, je1-js1+1, ke1-ks1+1, 2*nm0 
    call mpi_finalize(merr)
    stop        
  endif
!
  write(*,*) 'is,ie,js,je,ks,ke=',is,ie,js,je,ks,ke 
!
!======================================================================@
!  Allocate variables
!======================================================================@
!
  allocate( uu(nv,is1:ie1,js1:je1,ks1:ke1), &
            uh(nv,is1:ie1,js1:je1,ks1:ke1), &
            uo(nv,is1:ie1,js1:je1,ks1:ke1), &
            u0(nv,is1:ie1,js1:je1,ks1:ke1), &
            uusr(nv,is1:ie1,js1:je1,ks1:ke1), &
            uuir(nv,is1:ie1,js1:je1,ks1:ke1), &
            uuil(nv,is1:ie1,js1:je1,ks1:ke1), &
            uujr(nv,is1:ie1,js1:je1,ks1:ke1), &
            uujl(nv,is1:ie1,js1:je1,ks1:ke1), &
            uukr(nv,is1:ie1,js1:je1,ks1:ke1), &
            uukl(nv,is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( uri(ns,is1:ie1,js1:je1,ks1:ke1), &
            urio(ns,is1:ie1,js1:je1,ks1:ke1), &
            uri0(ns,is1:ie1,js1:je1,ks1:ke1), &
            urisr(ns,is1:ie1,js1:je1,ks1:ke1), &
            uriir(ns,is1:ie1,js1:je1,ks1:ke1), &
            uriil(ns,is1:ie1,js1:je1,ks1:ke1), &
            urijr(ns,is1:ie1,js1:je1,ks1:ke1), &
            urijl(ns,is1:ie1,js1:je1,ks1:ke1), &
            urikr(ns,is1:ie1,js1:je1,ks1:ke1), &
            urikl(ns,is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( hh(0:3,is1:ie1,js1:je1,ks1:ke1), &
            hhi(0:3,is1:ie1,js1:je1,ks1:ke1), &
            hhj(0:3,is1:ie1,js1:je1,ks1:ke1), &
            hhk(0:3,is1:ie1,js1:je1,ks1:ke1), &
            hi(3,nv,is1:ie1,js1:je1,ks1:ke1), &
            hii(nv,is1:ie1,js1:je1,ks1:ke1), &
            hij(nv,is1:ie1,js1:je1,ks1:ke1), &
            hik(nv,is1:ie1,js1:je1,ks1:ke1), &
            ho(3,nv,is1:ie1,js1:je1,ks1:ke1), &
            hoi(nv,is1:ie1,js1:je1,ks1:ke1), &
            hoj(nv,is1:ie1,js1:je1,ks1:ke1), &
            hok(nv,is1:ie1,js1:je1,ks1:ke1), &
            gm(0:3,3,is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( ww(3,nv,is1:ie1,js1:je1,ks1:ke1), &
            wwo(3,nv,is1:ie1,js1:je1,ks1:ke1), &
            wwir(nv,is1:ie1,js1:je1,ks1:ke1), &
            wwil(nv,is1:ie1,js1:je1,ks1:ke1), &
            wwjr(nv,is1:ie1,js1:je1,ks1:ke1), &
            wwjl(nv,is1:ie1,js1:je1,ks1:ke1), &
            wwkr(nv,is1:ie1,js1:je1,ks1:ke1), &
            wwkl(nv,is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( sf(nv,is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( cmaxi(is1:ie1,js1:je1,ks1:ke1), &
            cmini(is1:ie1,js1:je1,ks1:ke1), &
            cmaxj(is1:ie1,js1:je1,ks1:ke1), &
            cminj(is1:ie1,js1:je1,ks1:ke1), &
            cmaxk(is1:ie1,js1:je1,ks1:ke1), &
            cmink(is1:ie1,js1:je1,ks1:ke1), stat=merr )

  allocate( b1(is1:ie1,js1:je1,ks1:ke1), b2(is1:ie1,js1:je1,ks1:ke1), &
            b3(is1:ie1,js1:je1,ks1:ke1), &
            v1(is1:ie1,js1:je1,ks1:ke1), v2(is1:ie1,js1:je1,ks1:ke1), &
            v3(is1:ie1,js1:je1,ks1:ke1), &
            de(is1:ie1,js1:je1,ks1:ke1), pr(is1:ie1,js1:je1,ks1:ke1), &
            e1(is1:ie1,js1:je1,ks1:ke1), e2(is1:ie1,js1:je1,ks1:ke1), &
            e3(is1:ie1,js1:je1,ks1:ke1), &
            ch(is1:ie1,js1:je1,ks1:ke1), sg(is1:ie1,js1:je1,ks1:ke1), &
            stat=merr )
! for GLM
  allocate( ps(is1:ie1,js1:je1,ks1:ke1), ph(is1:ie1,js1:je1,ks1:ke1), &
            stat=merr)

! for GLM
!
!======================================================================@
!    file open for restart
!======================================================================@
!
!    open( unit=1, file='ipram', status='old')
  write(filename1,991) myrank
  open( unit=9, file=filename1, status='unknown', &
        form='unformatted')
!
 991 format ('restart',i3.3,'.outdat')
!
!======================================================================@
!     Initialization of varibles
!======================================================================@
!
!     Fixed parameter
!
  istop=0
  ierror=0
!
  nd=0
!
  dt=0.01d0
  dtshot=tmax/float(nshot)
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        do m=1,nv
           ww(1,m,i,j,k)=0.d0
           ww(2,m,i,j,k)=0.d0
           ww(3,m,i,j,k)=0.d0

           uu(m,i,j,k)=0.d0
           uh(m,i,j,k)=0.d0         
           uuir(m,i,j,k)=0.d0
           uujr(m,i,j,k)=0.d0
           uukr(m,i,j,k)=0.d0
           uuil(m,i,j,k)=0.d0
           uujl(m,i,j,k)=0.d0
           uukl(m,i,j,k)=0.d0
         
           wwir(m,i,j,k)=0.d0
           wwjr(m,i,j,k)=0.d0
           wwkr(m,i,j,k)=0.d0
           wwil(m,i,j,k)=0.d0
           wwjl(m,i,j,k)=0.d0
           wwkl(m,i,j,k)=0.d0         

           sf(m,i,j,k)=0.d0
        enddo
      enddo
    enddo
  enddo

  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        do m=1,ns
           uri(m,i,j,k)=0.d0
           uriir(m,i,j,k)=0.d0
           urijr(m,i,j,k)=0.d0
           urikr(m,i,j,k)=0.d0
           uriil(m,i,j,k)=0.d0
           urijl(m,i,j,k)=0.d0
           urikl(m,i,j,k)=0.d0
        enddo
      enddo
    enddo
  enddo
!
!======================================================================@
!      Print list of parameters
!======================================================================@
!
  if(myrank .eq. 0) then
    write(6,*) 'job name: ',jname,',  date: ',date
    write(6,*) 'Restart : ',icres
    write(6,*) '         List of Parameters '
    write(6,*) 'dtime =',dt,',  tmax =', tmax
    write(6,*) 'itmax =',itmax
    write(6,*) 'imax  =',imax,',  jmax =',jmax,',  kmax =',kmax
    write(6,*) 'npe =',npe
    write(6,*) 'xmin  =',xmin,',  xmax =', xmax
    write(6,*) 'ymin  =',ymin,',  ymax =', ymax
    write(6,*) 'zmin  =',zmin,',  zmax =', zmax
    write(6,*) 'model =',model,',  metric =',metric
    write(6,*) 'ieos =',ieos
    write(6,*) 'c0 =',c0
    write(6,*) 'gamma =',gam,',  gam0 =',gam0
    write(6,*) 'irec =',irec,', icha =',icha
    write(6,*) 'ihll =',ihll,', iwvec =',iwvec
  endif
!
!======================================================================@
!     Grid generation
!======================================================================@
!
  call grid(x1,x2,x3,dx1,dx2,dx3,dx1a,dx2a,dx3a, &
            dx1b,dx2b,dx3b,x1a,x2a,x3a)
!
!======================================================================@
!     Initial setting for Metric (coordinates)
!======================================================================@
!
  if( metric .eq. 1 ) then
    call carmet(hh,hhi,hhj,hhk,is1,ie1,js1,je1,ks1,ke1)
    call cargeo(gm,is1,ie1,js1,je1,ks1,ke1)
  elseif( metric .eq. 2 ) then
    call cylmet(hh,x1,hhi,hhj,hhk,x1a,is1,ie1,js1,je1,ks1,ke1)
    call cylgeo(gm,hh,is1,ie1,js1,je1,ks1,ke1)
  elseif( metric .eq. 3 ) then
    call sphmet(hh,x1,x3,hhi,hhj,hhk,x1a,x3a,is1,ie1,js1,je1,ks1,ke1)
    call sphgeo(gm,x3,hh,is1,ie1,js1,je1,ks1,ke1)
  else
    stop
  endif
!
  call calhoi(hh,hi,ho,hhi,hhj,hhk,hii,hij,hik,hoi,hoj,hok, &
              is1,ie1,js1,je1,ks1,ke1)
!
!     Calculation of Kolmogorov-like Power Spectrum
!
  if(model .eq. 54 .or. model .eq. 56 .or. model .eq. 57) then
    cl=1.d0
    amdamax=zmax
!    amdamin=10.0*max(dx1a,dx3a)
    amdamin=0.01*zmax
    call kolspec(wkn,amdamax,amdamin,cl,pkn1d,pkn2d,pkn3d, &
                 totpkn1d,totpkn2d,totpkn3d,thetan,phin,thetan1,phin1, &
                 npe,myrank)    
  endif
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
!
!    Researrt simulation or Not
!
  if( icres .eq. 0 ) then
!
!      Initial Model Setup
!
    if(model .eq. 20) then
      call mdjetpro(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                    rj,vj,etaj,betaj,cmac,imagj,is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 21) then
      call mdprejet(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                    e0,rj,vj,vpre1,vplu1,opre1,is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 22) then
      call mdkinkjet(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     e0,rj,vj,nm0,is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 24) then
      call mdkinkjet2(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                      e0,rj,vj,is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 25) then
      call mdkinkjet3(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                      is1,ie1,js1,je1,ks1,ke1,npe,myrank)
    elseif(model .eq. 28) then
      call mdkinkjet4(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                      is1,ie1,js1,je1,ks1,ke1,npe,myrank)     
    elseif(model .eq. 23) then
      call md2dblast(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 26) then
      call md2dmagadv(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                      is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 29) then
      call md3dmagadv(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                      is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 27) then
      call md2drotor(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 31) then
      call md2dkh(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                  is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 30) then
      call md3dkh(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                  is1,ie1,js1,je1,ks1,ke1,npe,myrank)
    elseif(model .eq. 32) then
      call md2drecon(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     is1,ie1,js1,je1,ks1,ke1,cuc,eta0)
    elseif(model .eq. 33) then
      call md2dtear(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                    is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 50) then
      call mdshoctub(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg, &
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 51) then
      call mdalfvenpro(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1, &
                       is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 52) then
      call mdshelpro(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 53) then
      call mdshref(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg, &
                   is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 54) then
      call md2dshref(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     vj,wkn,pkn2d,pkn1d, &
                     totpkn2d,totpkn1d,thetan,phin,cmac,betaj,bm,&
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 55) then
      call mdshpro(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg, &
                   is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 56) then
      call md2dshpro(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     vj,wkn,pkn2d,pkn1d, &
                     totpkn2d,totpkn1d,thetan,phin,betaj,bm, &
                     is1,ie1,js1,je1,ks1,ke1)
    elseif(model .eq. 57) then
      call md3dshcol(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1,x2,x3, &
                     wkn,pkn2d,pkn1d, &
                     totpkn2d,totpkn1d,thetan,phin,thetan1,phin1,&
                     is1,ie1,js1,je1,ks1,ke1)    
    elseif(model .eq. 60) then
      call mdslfsht(de,pr,v1,v2,v3,b1,b2,b3,e1,e2,e3,ch,sg,x1, &
                    is1,ie1,js1,je1,ks1,ke1)
    else
      stop
    endif
!
   call setphs(ph,ps,is1,ie1,js1,je1,ks1,ke1)
!
    do k=ks1,ke1
      do j=js1,je1
        do i=is1,ie1
!
          uri(1,i,j,k)=de(i,j,k)
          uri(2,i,j,k)=v1(i,j,k)
          uri(3,i,j,k)=v2(i,j,k)
          uri(4,i,j,k)=v3(i,j,k)
          uri(5,i,j,k)=pr(i,j,k)
          uri(6,i,j,k)=0.d0
          uri(7,i,j,k)=b1(i,j,k)
          uri(8,i,j,k)=b2(i,j,k)
          uri(9,i,j,k)=b3(i,j,k)
          uri(10,i,j,k)=e1(i,j,k)
          uri(11,i,j,k)=e2(i,j,k)
          uri(12,i,j,k)=e3(i,j,k)
          uri(13,i,j,k)=ch(i,j,k)
          uri(14,i,j,k)=ph(i,j,k)
          uri(15,i,j,k)=ps(i,j,k)
          uri(16,i,j,k)=sg(i,j,k)

        enddo
      enddo
    enddo
!
!         Conserved Variables Set (uu)
!
    call caluu1(uri,uu,x1,is1,ie1,js1,je1,ks1,ke1)
!
!     Transfer of variables as initial value
!
    call ident(uu,uo,is1,ie1,js1,je1,ks1,ke1)
    call ident(uu,u0,is1,ie1,js1,je1,ks1,ke1)
    call ident(uu,uusr,is1,ie1,js1,je1,ks1,ke1)
    call ident1(uri,urio,is1,ie1,js1,je1,ks1,ke1)
    call ident1(uri,uri0,is1,ie1,js1,je1,ks1,ke1)
    call ident1(uri,urisr,is1,ie1,js1,je1,ks1,ke1)
!
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
!
!     Initialize for time advance loop start
!
    it0=0
    ihi=0
    time=0.d0
    timeh=0.d0
    tnext=0.d0
  endif
!
! come back point if icres .eq. 1
!
!======================================================================@
!     Startup for restart simulation
!======================================================================@
!
!  If restart is 'on' skip the initial setup and move this line
!
  if( icres .ne. 0 ) then
!    call restar(uu,u0,uri,uri0,it0,ihi,timeh,ks1,ke1)
    call restar2(uu,u0,uri,uri0,it0,ihi,timeh,nd,npe,myrank,&
                 is1,ie1,js1,je1,ks1,ke1)
!
!!      tnext=timeh+dtshot                               1997.12.04(Thu)
    tnext=(int(timeh/dtshot)+1)*dtshot
!
    call ident(uu,uusr,is1,ie1,js1,je1,ks1,ke1)
    call ident1(uri,urisr,is1,ie1,js1,je1,ks1,ke1)
  endif
!
!**********************************************************************@
!     !! Time Advance Step Loop Start !!
!**********************************************************************@
! 
  its=(itmax-it0+1)/nshot
  if( its .lt. 1 ) then
    its = 1
  endif
  itsr=it0
  ihisr=ihi
  isr=0
  timesr=timeh
!
  do it=it0,itmax
!
!-----------------------------------------------------------------------
!     CPU Time Check                                1995.04.05
!-----------------------------------------------------------------------
!
!    call kclock(ig)
    if(myrank .eq. 0) then
      call cpu_time(cputime)
    endif
!
    call mpi_barrier(mpi_comm_world,merr)
    call mpi_bcast(cputime,1,mpi_double_precision,0,mpi_comm_world,merr)
!
!    if( ig .ge. icpu ) then
    if(cputime .ge. float(icpu)) then
      if(myrank .eq. 0) then
        write(6,*) ' CPU Time Over : cputime =',cputime,'(sec)'
       endif
       istop=1
    endif
!
    iit=it-it0
!
!x   if( mod(it,its) .eq. 0  .or. istop.eq.1 .or. it.eq.itmax ) then
!
!
!----------------------------------------------------------------------
!     Simulation time check
!----------------------------------------------------------------------
!
    if( dtmin.ge.0.d0 ) then
      timeo=time
      time=timeh
    else
      time = dt*float(it)
    endif
!
    if( time.gt.tmax ) then
      if(myrank .eq. 0) then
        write(6,*) '  '
        write(6,*) 'Congraturations! Run Succeeded! : time =' &
                    ,time,', tmax',tmax
        write(6,*) '  '
      endif
      istop = 1
    endif
!
!----------------------------------------------------------------------
!     Data store and output
!----------------------------------------------------------------------
!
    call mpi_barrier(mpi_comm_world,merr)
!
    if( time.ge.tnext .or. ierror.ne.0 &
        .or. istop.eq.1 .or. it.eq.itmax ) then
!
      if( ierror.eq.1 ) then
        write(6,*) 'stop by fault; time =',time
      endif
!
      ihi=ihi+it-it0+1
      if(time .eq. 0.0d0) then
       nd=0
      else
       nd=nd+1
      endif
!
! Data store for restart
!
      if( (isskip.eq.0 .or. istop.eq.1) .and. ierror.eq.0 ) then
        call store1(uu,u0,uri,uri0,it,ihi,time,nd,npe,myrank,&
                    is1,ie1,js1,je1,ks1,ke1)
      elseif( it.eq.itmax .and. ierror.eq.0 ) then
        call store1(uu,u0,uri,uri0,it,ihi,time,nd,npe,myrank,&
                    is1,ie1,js1,je1,ks1,ke1)
      elseif( mod(isr,isskip).eq.0 .and. ierror.eq.0 ) then
        call store1(uusr,u0,urisr,uri0,itsr,ihisr,timesr,nd,npe,myrank, &
                    is1,ie1,js1,je1,ks1,ke1)
        itsr=it
        ihisr=ihi
        timesr=time
        call ident(uu,uusr,is1,ie1,js1,je1,ks1,ke1)
        call ident1(uri,urisr,is1,ie1,js1,je1,ks1,ke1)
      endif
!
! Data output as Simulation results
!
      if(myrank .eq. 0) then
        write(6,*) ' data output for analysis'
        write(6,*) 'time=',time
!        write(6,*) 'dtshot=',dtshot
!        write(6,*) 'tnext=',tnext
        write(6,*) 'cputime=',cputime
      endif
!
      call output1(uri,x1,x2,x3,time,nm0,nd,npe,myrank,myranki,myrankj,&
                   myrankk,is1,ie1,js1,je1,ks1,ke1)
!
      isr=isr+1
      tnext=tnext+dtshot
!
    endif
!
    if( ierror .ne. 0 .or. istop .eq. 1 .or. it .eq. itmax)  then
      goto 100
    endif
!
!=======================================================================
!      Start next time step calculation
!=======================================================================
!
!-----------------------------------------------------------------------
!     Variable tansfer (before 1st step)
!-----------------------------------------------------------------------
!
    call ident(uu,uo,is1,ie1,js1,je1,ks1,ke1)
    call ident1(uri,urio,is1,ie1,js1,je1,ks1,ke1)
!
!!======================================================================@
!     Calcuration of dt for next time step
!======================================================================@
!
    if( dtmin.ge.0.d0 ) then

      call cdtcfl5(uri,dtcfl,dx1,dx2,dx3,dt1,dt2,dt3,nm0, &
                   is1,ie1,js1,je1,ks1,ke1)
!
! MPI
!
      call mpi_allreduce(dtcfl,dtcflg,1,mpi_double_precision,mpi_min, &
                         mpi_comm_world,merr)
      dt=cfl*dtcflg
!
      do i=1,imax-1
        akap1(i)=dt/dx1(i)
        akap1b(i)=dt/dx1b(i)
      enddo
      akap1a=dt/dx1a
!
      do j=1,jmax-1
        akap2(j)=dt/dx2(j)
        akap2b(j)=dt/dx2b(j)
      enddo
      akap2a=dt/dx2a
!
      do k=1,kmax-1
        akap3(k)=dt/dx3(k)
        akap3b(k)=dt/dx3b(k)
      enddo
      akap3a=dt/dx3a
!
      timeh=time+dt
    else
      timeh= dt*(float(it)+1.0)
    endif
       
    if( dt.lt.dtmin ) then
!      write(6,*) ' dt < dtmin: dt =',dt,', dtmin =',dtmin
      istop = 1
    endif
!
!-----------------------------------------------------------------------
!
    if(myrank .eq. 0) then
      write(6,620) 'Simulation running: it = ',it, &
                 ',  dt =',dt,', time=',time
620   format(1h ,a27,i6, 2(a7,1pe11.3))
    endif
!
!**********************************************************************@
!  GLM Calculation
!**********************************************************************@
!
!  Wave speed c_h for GLM
!
    call calglmch1(glmch,glmcp,dt,dx1a,dx2a,dx3a)
!
!  Add 1/2 step advance of stiff part of source term    
!
    call relax(uri,uu,glmch,dt,is1,ie1,js1,je1,ks1,ke1)
!    call recovb(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
!
!    call mpi_barrier(mpi_comm_world,merr)
!    call mpiex1(uri,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
!               myrankk,icputable)
!    call mpiex(uu,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
!               myrankk,icputable)
!
!**********************************************************************@
!     First Step for hydro part
!**********************************************************************@
!
!     Reconstraction Step
!
    call rec2(uri,x1,x2,x3,dx1,dx2,dx3,dx1b,dx2b,dx3b, &
              uriir,urijr,urikr,uriil,urijl,urikl, &
              uuir,uujr,uukr,uuil,uujl,uukl, &
              wwir,wwjr,wwkr,wwil,wwjl,wwkl,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Calculation of Characteristics
!
    call calcha5a(cmaxi,cmini,cmaxj,cminj,cmaxk,cmink,nm0, &
                  is1,ie1,js1,je1,ks1,ke1)
!
!     Calculation of Numerical flux 
!
    call hll(ww,x1,x2,x3,uri,uriir,uriil,urijr,urijl,urikr,urikl, &
             uuir,uujr,uukr,uuil,uujl,uukl, &
             wwir,wwjr,wwkr,wwil,wwjl,wwkl, &
             cmaxi,cmini,cmaxj,cminj,cmaxk,cmink,nm0, &
             is1,ie1,js1,je1,ks1,ke1)
!
!   Calculation of Numerical flux for GLM
!
    call glmhll(ww,uuir,uujr,uukr,uuil,uujl,uukl, &
                wwir,wwjr,wwkr,wwil,wwjl,wwkl, &
               glmch,nm0,is1,ie1,js1,je1,ks1,ke1)
!
    call calwwo(uu,wwo,uri,is1,ie1,js1,je1,ks1,ke1)
!
!     Calcuration of Source term
!
    call calsf(sf,uu,wwo,uri,gm,dt,is1,ie1,js1,je1,ks1,ke1)
!
!     Time advance of first step
!
    call rk2fst(uu,uh,ww,hi,ho,hii,hij,hik,hoi,hoj,hok,nm0, &
                akap1b,akap2b,akap3b,is1,ie1,js1,je1,ks1,ke1)
!
!     Add Source term
!
    call rk2adsff(uh,sf,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Boundary condition for conserved variables
!
!y   call bnd4(uu,uo,x1,nm0,is1,ie1,js1,je1,ks1,ke1, &
!y             myranki,myrankj,myrankk,icputable)
!
!     Recovery step
!
!    call mpi_barrier(mpi_comm_world,merr)
    call recov(uh,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Boundary condition for primitive variables
!
!    call mpi_barrier(mpi_comm_world,merr)
    call bnd4a(uri,urio,x1,nm0,is1,ie1,js1,je1,ks1,ke1, &
               myranki,myrankj,myrankk,icputable)
!
!     Calculation of conserved variables at boundary region
!
!
!    call mpi_barrier(mpi_comm_world,merr)
    call calconv(uh,uri,x1,nm0,is1,ie1,js1,je1,ks1,ke1, &
                 myranki,myrankj,myrankk)
!
!      MPI data exchange
!
    call mpi_barrier(mpi_comm_world,merr)
    call mpiex1(uri,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
               myrankk,icputable)
    call mpi_barrier(mpi_comm_world,merr)
    call mpiex(uh,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
               myrankk,icputable)
!
!     Check and Correction (velocity)
!
    if( adampa.ne.0.d0 ) then
      call spdlia1(uri,uh,istop,ierror,is1,ie1,js1,je1,ks1,ke1)
    endif
!
!     Check and Correction (pressure and density)
!
    call pminmax1a(uh,uri,is1,ie1,js1,je1,ks1,ke1)
!
!     Artificial Damping (primitive variables) 
!
    if(model .ne. 22) then
      if(adamp.gt.0.d0 .and. rdamp.gt.0.d0) then
        call damp4(uh,uri,uri0,x1,x3,is1,ie1,js1,je1,ks1,ke1)
      endif
    elseif(model .eq. 22) then
      if(adamp.gt.0.d0) then
        call damp4a(uh,uri,uri0,x1,x3,is1,ie1,js1,je1,ks1,ke1)
      endif
    endif
!
!     Input Jet from boundary (option)
!
!    if(model .eq. 20) then        
!      call inpjet3(uh,uri,x1,x2,x3,rj,vj,etaj,betaj,cmac,imagj,&
!                   is1,ie1,js1,je1,ks1,ke1)
!    endif
!
!     Input Precession for Jet Propagation Simulation (option)
!
    if(model .eq. 21) then
      time1=time+0.5*dt
      call prece1(uh,uri,uri0,x1,x2,x3,rj,vj,vpre1,vplu1,opre1,time1,&
                  is1,ie1,js1,je1,ks1,ke1)
    endif
!
!     Input Rotation for 3D kink Instability Simulation (option)
!
!   if(model .eq. 22) then
!     call rotate1(uh,uri,uri0,x1,x2,x3,omk1,nm0)
!   endif
!
!     Input unshocked-flow from boundary (option)
!
    if(model .eq. 54) then
!      call cleanb(uh,uri,x1,x2,x3,vj,cmac,betaj,bm)
      time1=time+0.5*dt
      call inpflow1(uh,uri,x1,x2,x3, &
                    vj,wkn,pkn2d,pkn1d,totpkn2d,totpkn1d, &
                    thetan,phin,cmac,betaj,bm,time1,it, &
                    is1,ie1,js1,je1,ks1,ke1)
!      call smooth(uh,uri,x1,x2,x3,time1)
    endif
!
!**********************************************************************@
!     Second Step
!**********************************************************************@
!
!     Reconstraction Step
!
    call rec2(uri,x1,x2,x3,dx1,dx2,dx3,dx1b,dx2b,dx3b, &
              uriir,urijr,urikr,uriil,urijl,urikl, &
              uuir,uujr,uukr,uuil,uujl,uukl, &
              wwir,wwjr,wwkr,wwil,wwjl,wwkl,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Calculation of Characteristics
!
    call calcha5a(cmaxi,cmini,cmaxj,cminj,cmaxk,cmink,nm0, &
                  is1,ie1,js1,je1,ks1,ke1)
!
!     Calculation of Numerical flux 
!
    call hll(ww,x1,x2,x3,uri,uriir,uriil,urijr,urijl,urikr,urikl, &
             uuir,uujr,uukr,uuil,uujl,uukl, &
             wwir,wwjr,wwkr,wwil,wwjl,wwkl, &
             cmaxi,cmini,cmaxj,cminj,cmaxk,cmink,nm0, &
             is1,ie1,js1,je1,ks1,ke1)
!
!   Calculation of Numerical flux for GLM
!
    call glmhll(ww,uuir,uujr,uukr,uuil,uujl,uukl, &
                wwir,wwjr,wwkr,wwil,wwjl,wwkl, &
                glmch,nm0,is1,ie1,js1,je1,ks1,ke1)
!
    call calwwo(uh,wwo,uri,is1,ie1,js1,je1,ks1,ke1)
!
!     Calcuration of source term 
!
    call calsf(sf,uh,wwo,uri,gm,dt,is1,ie1,js1,je1,ks1,ke1)
!
!     Time advance of second step
!
    call rk2snd(uu,ww,hi,ho,hii,hij,hik,hoi,hoj,hok,nm0, &
                akap1b,akap2b,akap3b,is1,ie1,js1,je1,ks1,ke1)
!
!     Add the Source Term
!
    call rk2adsfs(uu,sf,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Boundary Condition for conserved variables
!
!y   call bnd4(uu,uo,x1,nm0,is1,ie1,js1,je1,ks1,ke1,&
!y             myranki,myrankj,myrankk,icputable)
!
!  Add 2/2 step advance of stiff part of source term    
!
    call relaxb(uri,uu,glmch,dt,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Recovery step
!
    call recov(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!
!     Boundary Condition for primitive variables
!
    call bnd4a(uri,urio,x1,nm0,is1,ie1,js1,je1,ks1,ke1,&
               myranki,myrankj,myrankk,icputable)
!
!     Calculation of conserved variables at boundary region
!
    call calconv(uu,uri,x1,nm0,is1,ie1,js1,je1,ks1,ke1, &
                 myranki,myrankj,myrankk)
!
!      Calculation of new resistivity (current-dependent)
!
!    call calsg(uri,cuc,eta0,is1,ie1,js1,je1,ks1,ke1)
!
!      MPI data exchange
!
    call mpi_barrier(mpi_comm_world,merr)
    call mpiex1(uri,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
               myrankk,icputable)
    call mpi_barrier(mpi_comm_world,merr)
    call mpiex(uu,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
               myrankk,icputable)
!
!     Check and Correction (velocity)
!
    if( adampa.ne.0.d0 ) then
      call spdlia1(uri,uu,istop,ierror,is1,ie1,js1,je1,ks1,ke1)
    endif
!
!     Check and Correction (pressure and density)
!
    call pminmax1a(uu,uri,is1,ie1,js1,je1,ks1,ke1)
!
!     Artificial Damping (primitive variables) 
!
    if(model .ne. 22) then
      if(adamp.gt.0.d0 .and. rdamp.gt.0.d0) then
        call damp4(uu,uri,uri0,x1,x3,is1,ie1,js1,je1,ks1,ke1)
      endif
    elseif(model .eq. 22) then
      if(adamp.gt.0.d0) then
        call damp4a(uu,uri,uri0,x1,x3,is1,ie1,js1,je1,ks1,ke1)
      endif
    endif
!
!     Input Jet from boundary (option)
!
    if(model .eq. 20) then        
      call inpjet3(uu,uri,x1,x2,x3,rj,vj,etaj,betaj,cmac,imagj,&
                   is1,ie1,js1,je1,ks1,ke1)
    endif
!
!     Input Precession for Jet Propagation Simulation (option)
!
    if(model .eq. 21) then
      time1=time+dt
      call prece1(uu,uri,uri0,x1,x2,x3,rj,vj,vpre1,vplu1,opre1,time1,&
                  is1,ie1,js1,je1,ks1,ke1)
    endif
!
!     Input Rotation for 3D kink Instability Simulation (option)
!
!    if(model .eq. 22) then
!      call rotate1(uu,uri,uri0,x1,x2,x3,rj,omk1,nm0)
!    endif
!
!     Input unshocked-flow from boundary (option)
!
    if(model .eq. 54) then
!      call cleanb(uu,uri,x1,x2,x3,vj,cmac,betaj,bm)        
      time1=time+dt
      call inpflow1(uu,uri,x1,x2,x3, &
                    vj,wkn,pkn2d,pkn1d,totpkn2d,totpkn1d, &
                    thetan,phin,cmac,betaj,bm,time1,it, &
                    is1,ie1,js1,je1,ks1,ke1)
!      call smooth(uu,uri,x1,x2,x3,time1)
    endif
!!
!**********************************************************************@
!  Calcualtion for GLM 
!**********************************************************************@
!  Add 2/2 step advance of stiff part of source term    
!
!    call relax(uri,uu,glmch,dt,is1,ie1,js1,je1,ks1,ke1)
!    call recovb(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
!
!    call mpi_barrier(mpi_comm_world,merr)
!    call mpiex1(uri,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
!               myrankk,icputable)
!    call mpi_barrier(mpi_comm_world,merr)
!    call mpiex(uu,nm0,is1,ie1,js1,je1,ks1,ke1,myranki,myrankj,&
!               myrankk,icputable)
!
!ccc
  enddo
!
!**********************************************************************@
!     !! Time Step Loop End !!
!**********************************************************************@
!
100 continue

  call mpi_finalize(merr)

  if(myrank.eq. 0) then
    write(6,*) '== end =='
  endif

  rewind(8)
!
!======================================================================@
!     CPU Time for End of Job
!======================================================================@
!
!y  it0=it
!y  call kclock(ig)
!y  write(6,*) 'CPU Time at End of Job :',ig,'sec'
!   .
!======================================================================@
  stop
end program main
